/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2022 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2022 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/acm/permission_rule.h"
#include "caosdb/acm/permission_rule_impl.h" // for PermissionRuleImpl
#include "caosdb/acm/v1alpha1/main.pb.h"     // for ListPermissionRulesRequest
#include "caosdb/protobuf_helper.h"          // for ProtoMessageWrapper
#include <type_traits>                       // for remove_reference<>::type
#include <utility>                           // for move

namespace caosdb::acm {
using caosdb::utility::ScalarProtoMessageWrapper;
using ProtoPermissionRule = caosdb::acm::v1alpha1::PermissionRule;

PermissionRuleImpl::PermissionRuleImpl() = default;

PermissionRuleImpl::PermissionRuleImpl(ProtoPermissionRule *rule)
  : ScalarProtoMessageWrapper<ProtoPermissionRule>(rule) {}

PermissionRuleImpl::PermissionRuleImpl(std::string permission, bool isGrant, bool isPriority) {
  this->wrapped->set_permission(std::move(permission));
  this->wrapped->set_grant(isGrant);
  this->wrapped->set_priority(isPriority);
}

PermissionRule::PermissionRule(std::string permission, bool isGrant, bool isPriority)
  : impl(new PermissionRuleImpl(std::move(permission), isGrant, isPriority)) {}

PermissionRule::PermissionRule(const PermissionRule &rule) : impl(new PermissionRuleImpl()) {
  this->impl->wrapped->CopyFrom(*(rule.impl->wrapped));
}

auto PermissionRule::operator=(const PermissionRule &rule) -> PermissionRule & {
  if (this == &rule) {
    return *this;
  }
  this->impl->wrapped->CopyFrom(*(rule.impl->wrapped));

  return *this;
}

PermissionRule::~PermissionRule() { delete this->impl; }

auto PermissionRule::ToString() const -> std::string { return this->impl->ToString(); }

[[nodiscard]] auto PermissionRule::IsGrant() const -> bool { return this->impl->wrapped->grant(); }

auto PermissionRule::SetGrant(bool isGrant) -> void { this->impl->wrapped->set_grant(isGrant); }

[[nodiscard]] auto PermissionRule::IsPriority() const -> bool {
  return this->impl->wrapped->priority();
}

auto PermissionRule::SetPriority(bool isPriority) -> void {
  this->impl->wrapped->set_priority(isPriority);
}

auto PermissionRule::operator==(const PermissionRule &other) const -> bool {
  if (this == &other) {
    return true;
  }
  return this->GetPermission() == other.GetPermission() &&
         this->IsPriority() == other.IsPriority() && this->IsGrant() == other.IsGrant();
}

auto HashPermissionRule::operator()(const PermissionRule &rule) const -> size_t {
  return std::hash<std::string>()(rule.GetPermission());
}

auto PermissionRule::GetPermission() const -> const std::string & {
  return this->impl->wrapped->permission();
}

} // namespace caosdb::acm
