/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2022 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2022 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/acm/role.h"
#include "caosdb/acm/permission_rule.h"  // for PermissionRule
#include "caosdb/acm/role_impl.h"        // for RoleImpl
#include "caosdb/acm/v1alpha1/main.pb.h" // for ListRolesRequest
#include "caosdb/protobuf_helper.h"      // for ProtoMessageWrapper
#include <type_traits>                   // for remove_reference<>::type
#include <unordered_set>                 // for _Node_iterator, operator!=
#include <utility>                       // for move
// IWYU pragma: no_include "net/proto2/public/repeated_field.h"

namespace caosdb::acm {
using caosdb::utility::ScalarProtoMessageWrapper;
using ProtoRole = caosdb::acm::v1alpha1::Role;
using ProtoRolePermissions = caosdb::acm::v1alpha1::RolePermissions;
using ProtoRoleCapabilities = caosdb::acm::v1alpha1::RoleCapabilities;
using ProtoListRoleItem = caosdb::acm::v1alpha1::ListRoleItem;

RoleImpl::RoleImpl(std::string name, std::string description) {
  if (!name.empty()) {
    this->wrapped->set_name(name);
  }
  if (!description.empty()) {
    this->wrapped->set_description(description);
  }
}

RoleImpl::RoleImpl(ProtoRole *role) : ScalarProtoMessageWrapper<ProtoRole>(role) {}

RoleImpl::RoleImpl(ProtoListRoleItem &role_item)
  : ScalarProtoMessageWrapper<ProtoRole>(role_item.release_role()) {}

RoleImpl::RoleImpl(std::string name) : RoleImpl(std::move(name), "") {}

[[nodiscard]] auto RoleImpl::GeneratePermissionRulesSet() -> std::unique_ptr<PermissionRules> {
  auto result = std::make_unique<PermissionRules>();
  for (auto const &rule : this->wrapped->permission_rules()) {
    result->emplace(rule.permission(), rule.grant(), rule.priority());
  }
  return result;
}

Role::Role(std::unique_ptr<RoleImpl> wrapped) : wrapped(std::move(wrapped)) {}

Role::Role(std::string name, std::string description)
  : wrapped(std::make_unique<RoleImpl>(std::move(name), std::move(description))) {}

Role::Role(std::string name) : Role(std::move(name), {""}) {}

Role::Role(const Role &role) : Role(role.GetName(), role.GetDescription()) {}

auto Role::operator=(const Role &role) -> Role & {
  if (this == &role) {
    return *this;
  }
  this->wrapped->wrapped->CopyFrom(*(role.wrapped->wrapped));

  return *this;
}

Role::~Role() = default;
auto Role::GetName() const -> const std::string & { return this->wrapped->wrapped->name(); }

auto Role::SetName(std::string name) -> void {
  if (!name.empty()) {
    this->wrapped->wrapped->set_name(std::move(name));
  } else {
    this->wrapped->wrapped->clear_name();
  }
}

auto Role::GetDescription() const -> const std::string & {
  return this->wrapped->wrapped->description();
}

auto Role::SetDescription(std::string description) -> void {
  if (!description.empty()) {
    this->wrapped->wrapped->set_description(std::move(description));
  } else {
    this->wrapped->wrapped->clear_description();
  }
}

auto Role::ToString() const -> std::string { return this->wrapped->ToString(); }

[[nodiscard]] auto Role::GetPermissionRules() const -> const PermissionRules & {
  if (this->wrapped->rules == nullptr) {
    this->wrapped->rules = this->wrapped->GeneratePermissionRulesSet();
  }
  return *(this->wrapped->rules);
}

auto Role::Grant(const std::string &permission, bool isPriority) -> void {
  this->AddPermissionRule(permission, true, isPriority);
}

auto Role::RevokeGrant(const std::string &permission, bool isPriority) -> void {
  this->RemovePermissionRule(permission, true, isPriority);
}

auto Role::Deny(const std::string &permission, bool isPriority) -> void {
  this->AddPermissionRule(permission, false, isPriority);
}

auto Role::RevokeDenial(const std::string &permission, bool isPriority) -> void {
  this->RemovePermissionRule(permission, false, isPriority);
}

auto Role::AddPermissionRule(const std::string &permission, bool isGrant, bool isPriority) -> void {
  this->RemovePermissionRule(permission, isGrant, isPriority);
  auto *new_rule = this->wrapped->wrapped->add_permission_rules();
  new_rule->set_permission(permission);
  new_rule->set_priority(isPriority);
  new_rule->set_grant(isGrant);
}

auto Role::RemovePermissionRule(const std::string &permission, bool isGrant, bool isPriority)
  -> void {
  this->wrapped->rules.reset();
  auto rule = this->wrapped->wrapped->mutable_permission_rules()->begin();
  while (rule != this->wrapped->wrapped->mutable_permission_rules()->end()) {
    if (rule->grant() == isGrant && rule->priority() == isPriority &&
        rule->permission() == permission) {
      this->wrapped->wrapped->mutable_permission_rules()->erase(rule);
      rule = this->wrapped->wrapped->mutable_permission_rules()->begin();
    } else {
      rule++;
    }
  }
}

auto Role::ClearPermissionRules() -> void {
  this->wrapped->rules.reset();
  this->wrapped->wrapped->clear_permission_rules();
}

} // namespace caosdb::acm
