/*
 *
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/connection.h"
#include "caosdb/acm/role_impl.h"             // for RoleImpl
#include "caosdb/acm/user_impl.h"             // for UserImpl
#include "caosdb/acm/v1alpha1/main.grpc.pb.h" // for AccessControlMan...
#include "caosdb/acm/v1alpha1/main.pb.h"      // for ListRolesRequest
#include "caosdb/configuration.h"             // for ConnectionConfigur...
#include "caosdb/exceptions.h"                // for ConfigurationError
#include "caosdb/info.h"                      // for VersionInfo
#include "caosdb/info/v1/main.grpc.pb.h"      // for GeneralInfoService
#include "caosdb/info/v1/main.pb.h"           // for GetVersionInfoRequest
#include "caosdb/transaction.h"               // for Transaction
#include "caosdb/transaction_status.h"        // for TransactionStatus
#include <grpcpp/client_context.h>            // for ClientContext
#include <grpcpp/create_channel.h>            // for CreateChannel
#include <grpcpp/support/status.h>            // for Status
#include <grpcpp/support/status_code_enum.h>  // for StatusCode, UNAUTHENTIC...
#include <string>                             // for string, operator+
// IWYU pragma: no_include "net/proto2/public/repeated_field.h"

namespace caosdb::connection {
using caosdb::acm::RoleImpl;
using caosdb::acm::UserImpl;
using caosdb::acm::v1alpha1::AccessControlManagementService;
using caosdb::acm::v1alpha1::CreateSingleRoleRequest;
using caosdb::acm::v1alpha1::CreateSingleRoleResponse;
using caosdb::acm::v1alpha1::CreateSingleUserRequest;
using caosdb::acm::v1alpha1::CreateSingleUserResponse;
using caosdb::acm::v1alpha1::DeleteSingleRoleRequest;
using caosdb::acm::v1alpha1::DeleteSingleRoleResponse;
using caosdb::acm::v1alpha1::DeleteSingleUserRequest;
using caosdb::acm::v1alpha1::DeleteSingleUserResponse;
using caosdb::acm::v1alpha1::ListRolesRequest;
using caosdb::acm::v1alpha1::ListRolesResponse;
using caosdb::acm::v1alpha1::RetrieveSingleRoleRequest;
using caosdb::acm::v1alpha1::RetrieveSingleRoleResponse;
using caosdb::acm::v1alpha1::RetrieveSingleUserRequest;
using caosdb::acm::v1alpha1::RetrieveSingleUserResponse;
using caosdb::configuration::ConfigurationManager;
using caosdb::configuration::ConnectionConfiguration;
using caosdb::entity::v1::EntityTransactionService;
using caosdb::entity::v1::FileTransmissionService;
using caosdb::info::VersionInfo;
using caosdb::info::v1::GeneralInfoService;
using caosdb::info::v1::GetVersionInfoRequest;
using caosdb::info::v1::GetVersionInfoResponse;
using caosdb::transaction::Transaction;
using caosdb::transaction::TransactionStatus;

Connection::Connection(const ConnectionConfiguration &configuration) {
  const std::string target =
    configuration.GetHost() + ":" + std::to_string(configuration.GetPort());
  this->channel = grpc::CreateChannel(target, configuration.GetChannelCredentials());
  this->general_info_service = GeneralInfoService::NewStub(this->channel);
  this->entity_transaction_service =
    std::make_shared<EntityTransactionService::Stub>(this->channel);
  this->file_transmission_service = std::make_shared<FileTransmissionService::Stub>(this->channel);
  this->access_controll_management_service =
    std::make_unique<AccessControlManagementService::Stub>(this->channel);
}

auto Connection::RetrieveVersionInfoNoExceptions() const noexcept -> TransactionStatus {

  const GetVersionInfoRequest request;
  GetVersionInfoResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->general_info_service->GetVersionInfo(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status = TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(std::to_string(grpc_status.error_code()) + " - " +
                                            error_message);
    }
  } else {
    this->version_info = std::make_unique<VersionInfo>(response.release_version_info());
  }

  return status;
}

auto Connection::RetrieveVersionInfo() const -> const VersionInfo & {

  TransactionStatus status = RetrieveVersionInfoNoExceptions();
  status.ThrowExceptionIfError();
  return *GetVersionInfo();
}

[[nodiscard]] auto Connection::CreateTransaction() const -> std::unique_ptr<Transaction> {
  auto entity_service = this->entity_transaction_service;
  auto file_service = this->file_transmission_service;
  return std::make_unique<Transaction>(entity_service, file_service);
}

[[nodiscard]] auto Connection::RetrieveSingleRole(std::string name) const -> Role {
  RetrieveSingleRoleRequest request;
  request.set_name(name);
  RetrieveSingleRoleResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->access_controll_management_service->RetrieveSingleRole(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status = TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(std::to_string(grpc_status.error_code()) + " - " +
                                            error_message);
    }
  }
  status.ThrowExceptionIfError();
  auto *role = response.release_role();
  return Role(std::make_unique<RoleImpl>(role));
}

auto Connection::DeleteSingleRole(std::string name) const -> void {
  DeleteSingleRoleRequest request;
  request.set_name(name);
  DeleteSingleRoleResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->access_controll_management_service->DeleteSingleRole(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  std::vector<Role> result;
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status = TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(std::to_string(grpc_status.error_code()) + " - " +
                                            error_message);
    }
  }
  status.ThrowExceptionIfError();
}

auto Connection::CreateSingleRole(const Role &role) const -> void {
  CreateSingleRoleRequest request;
  request.set_allocated_role(role.wrapped->wrapped);
  CreateSingleRoleResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->access_controll_management_service->CreateSingleRole(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status = TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(std::to_string(grpc_status.error_code()) + " - " +
                                            error_message);
    }
  }
  status.ThrowExceptionIfError();
}

// TODO(tf) find a way to deal with this:
// NOLINTNEXTLINE(bugprone-easily-swappable-parameters)
[[nodiscard]] auto Connection::RetrieveSingleUser(std::string realm, std::string name) const
  -> User {
  RetrieveSingleUserRequest request;
  request.set_name(name);
  request.set_realm(realm);
  RetrieveSingleUserResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->access_controll_management_service->RetrieveSingleUser(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status = TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(std::to_string(grpc_status.error_code()) + " - " +
                                            error_message);
    }
  }
  status.ThrowExceptionIfError();
  auto *user = response.release_user();
  return User(std::make_unique<UserImpl>(user));
}

// TODO(tf) find a way to deal with this:
// NOLINTNEXTLINE(bugprone-easily-swappable-parameters)
auto Connection::DeleteSingleUser(std::string realm, std::string name) const -> void {
  DeleteSingleUserRequest request;
  request.set_name(name);
  request.set_realm(realm);
  DeleteSingleUserResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->access_controll_management_service->DeleteSingleUser(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status = TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(std::to_string(grpc_status.error_code()) + " - " +
                                            error_message);
    }
  }
  status.ThrowExceptionIfError();
}

auto Connection::CreateSingleUser(const User &user) const -> void {
  CreateSingleUserRequest request;
  request.set_allocated_user(user.wrapped->wrapped);
  request.mutable_password_setting()->set_password(user.GetPassword());
  CreateSingleUserResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->access_controll_management_service->CreateSingleUser(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  std::vector<Role> result;
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status = TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(std::to_string(grpc_status.error_code()) + " - " +
                                            error_message);
    }
  }
  status.ThrowExceptionIfError();
}

[[nodiscard]] auto Connection::ListRoles() const -> std::vector<Role> {
  const ListRolesRequest request;
  ListRolesResponse response;
  grpc::ClientContext context;
  const grpc::Status grpc_status =
    this->access_controll_management_service->ListRoles(&context, request, &response);

  auto status = TransactionStatus::SUCCESS();
  std::vector<Role> result;
  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      status = TransactionStatus::AUTHENTICATION_ERROR(grpc_status.error_message());
      break;
    case grpc::StatusCode::UNAVAILABLE:
      status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_message = grpc_status.error_message();
      status = TransactionStatus::RPC_ERROR(std::to_string(grpc_status.error_code()) + " - " +
                                            error_message);
    }
  }
  status.ThrowExceptionIfError();

  auto *roles = response.mutable_roles();
  for (auto &role_item : *roles) {
    result.emplace_back(std::make_unique<RoleImpl>(role_item));
  }
  return result;
}

auto ConnectionManager::mHasConnection(const std::string &name) const -> bool {
  auto it = connections.find(name);
  return it != connections.end();
}

auto ConnectionManager::mGetConnection(const std::string &name) const
  -> const std::shared_ptr<Connection> & {
  if (!HasConnection(name)) {
    try {
      auto connection = ConfigurationManager::GetConnectionConfiguration(name);
      connections[name] = std::make_shared<Connection>(*connection.release());
    } catch (const caosdb::exceptions::ConfigurationError &exc) {
      throw caosdb::exceptions::ConnectionConfigurationError("Error with the connection named '" +
                                                             name + "': " + exc.what());
    }
  }
  return this->connections.at(name);
}

auto ConnectionManager::mGetDefaultConnection() const -> const std::shared_ptr<Connection> & {
  if (!HasConnection(default_connection_name)) {
    default_connection_name = ConfigurationManager::GetDefaultConnectionName();
    auto default_connection = ConfigurationManager::GetDefaultConnectionConfiguration();
    connections[default_connection_name] =
      std::make_shared<Connection>(*default_connection.release());
  }

  return connections.at(default_connection_name);
}

} // namespace caosdb::connection
