/*
 *
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

// A simple caosdb client
#include "caosdb/connection.h"         // for Connection, ConnectionManager
#include "caosdb/constants.h"          // for LIBCAOSDB_VERSION_MINOR, LIBCAOSDB_V...
#include "caosdb/entity.h"             // for Entity
#include "caosdb/exceptions.h"         // for ConfigurationError
#include "caosdb/info.h"               // for VersionInfo
#include "caosdb/logging.h"            // for CAOSDB_LOG_TRACE
#include "caosdb/transaction.h"        // for Transaction, ResultSet
#include "caosdb/transaction_status.h" // for TransactionSt...
#include <boost/program_options.hpp>   // for options_description
#include <exception>                   // for exception
#include <iostream>                    // for operator<<, basic_ostream, basic_ost...
#include <memory>                      // for unique_ptr, allocator, __shared_ptr_...
#include <string>                      // for operator<<, char_traits

const auto logger_name = "libcaosdb";

using boost::program_options::bool_switch;
using boost::program_options::command_line_parser;
using boost::program_options::notify;
using boost::program_options::options_description;
using boost::program_options::parse_command_line;
using boost::program_options::positional_options_description;
using boost::program_options::store;
using boost::program_options::value;
using boost::program_options::variables_map;
using caosdb::acm::Role;

auto print_version(bool print) -> void {
  if (print) {
    std::cout << "CaosDB C++ client (libcaosdb " << caosdb::LIBCAOSDB_VERSION_MAJOR << "."
              << caosdb::LIBCAOSDB_VERSION_MINOR << "." << caosdb::LIBCAOSDB_VERSION_PATCH << ")\n"
              << "We don't miss the H of caos.\n"
              << std::endl;
  }
}

auto test_connection() -> void {
  const auto &connection = caosdb::connection::ConnectionManager::GetDefaultConnection();

  connection->RetrieveVersionInfoNoExceptions();
  // get version info of the server
  const auto &v_info = connection->GetVersionInfo();
  std::cout << "Server Version: " << v_info->GetMajor() << "." << v_info->GetMinor() << "."
            << v_info->GetPatch() << "-" << v_info->GetPreRelease() << "-" << v_info->GetBuild()
            << std::endl;
}

auto retrieve_entity_by_id(std::string id) -> void {
  std::cout << "Retrieve entity " << id << std::endl;
  const auto &connection = caosdb::connection::ConnectionManager::GetDefaultConnection();
  auto transaction(connection->CreateTransaction());
  transaction->RetrieveById(id);
  transaction->ExecuteAsynchronously();
  auto t_stat = transaction->WaitForIt();
  CAOSDB_LOG_INFO(logger_name) << "status: " << t_stat.GetCode() << " // "
                               << t_stat.GetDescription();
  const auto &result_set = transaction->GetResultSet();
  for (const auto &entity : result_set) {
    std::cout << entity.ToString() << std::endl;
  }
}

auto execute_query(std::string query) -> void {
  std::cout << "Execute query:\n" << query << std::endl;
  const auto &connection = caosdb::connection::ConnectionManager::GetDefaultConnection();
  auto q_transaction(connection->CreateTransaction());
  q_transaction->Query(query);
  q_transaction->ExecuteAsynchronously();
  auto t_stat = q_transaction->WaitForIt();
  CAOSDB_LOG_INFO(logger_name) << "status: " << t_stat.GetCode() << " // "
                               << t_stat.GetDescription();
  const auto &result_set_2 = q_transaction->GetResultSet();
  for (const auto &entity : result_set_2) {
    std::cout << entity.ToString() << std::endl;
  }
}

auto list_roles() -> void {
  std::cout << "Known roles:" << std::endl;
  const auto &connection = caosdb::connection::ConnectionManager::GetDefaultConnection();
  auto roles = connection->ListRoles();
  for (const auto role : roles) {
    std::cout << role.ToString() << std::endl;
  }
}

auto retrieve_role(std::string name) {
  const auto &connection = caosdb::connection::ConnectionManager::GetDefaultConnection();
  auto role = connection->RetrieveSingleRole(name);
  std::cout << role.ToString() << std::endl;
}

auto create_role(std::string name) {
  const auto &connection = caosdb::connection::ConnectionManager::GetDefaultConnection();
  Role role(name);
  connection->CreateSingleRole(role);
  std::cout << "OK" << std::endl;
}

auto delete_role(std::string name) {
  const auto &connection = caosdb::connection::ConnectionManager::GetDefaultConnection();
  connection->DeleteSingleRole(name);
  std::cout << "OK" << std::endl;
}

auto main(int argc, const char *argv[]) -> int {
  try {
    command_line_parser parser(argc, argv);

    options_description desc{"Options"};
    desc.add_options()("help,h", "Show help.")("version,V", bool_switch()->notifier(print_version),
                                               "Print version of libcaosdb.")(
      "test-connection,t", "Test the connection to the caosdb server.")(
      "retrieve", value<std::string>()->notifier(retrieve_entity_by_id),
      "Retrieve an entity by id and print its JSON representation.")(
      "execute-query", value<std::string>()->notifier(execute_query),
      "Execute a query and print the results")("list-roles", "List all known roles")(
      "retrieve-role", value<std::string>()->notifier(retrieve_role), "Retrieve a role by name")(
      "create-role", value<std::string>()->notifier(create_role), "Create a new role")(
      "delete-role", value<std::string>()->notifier(delete_role), "Create a new role");

    parser.options(desc);

    // TODO(tf) add positional parameters
    // positional_options_description pos_desc;
    // pos_desc.add("phone", 1);
    // parser.positional(pos_desc);

    variables_map vm;
    store(parser.run(), vm);
    notify(vm);

    if (vm.count("help")) {
      std::cout << desc << std::endl;
    } else if (vm["version"].as<bool>()) {
    } else if (vm.count("test-connection")) {
      test_connection();
    } else if (vm.count("list-roles")) {
      list_roles();
    } else if (vm.count("retrieve-role")) {
    } else if (vm.count("retrieve")) {
    } else if (vm.count("create-role")) {
    } else if (vm.count("delete-role")) {
    } else if (vm.count("execute-query")) {
    } else {
      print_version(true);
      test_connection();
    }

    return 0;
  } catch (const caosdb::exceptions::ConfigurationError &exc) {
    std::cout << "ConfigurationError: " << exc.what() << std::endl;
    return exc.GetCode();
  } catch (const std::exception &exc) {
    std::cout << "Exception: " << exc.what() << std::endl;
    return 1;
  } catch (...) {
    std::cout << "Unknown error." << std::endl;
    return 2;
  }
}
