/*
 *
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2022 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2022 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "caosdb/acm/role.h"            // for Role
#include "caosdb/acm/permission_rule.h" // for PermissionRules, PermissionRule
#include <gtest/gtest-message.h>        // for Message
#include <gtest/gtest-test-part.h>      // for TestPartResult, SuiteApiRe...
#include <gtest/gtest_pred_impl.h>      // for Test, EXPECT_EQ, TEST
#include <string>                       // for allocator, string
#include <unordered_set>                // for _Node_const_iterator, operat...

namespace caosdb::acm {

TEST(test_role, role_name) {
  auto role = Role("role1");
  EXPECT_EQ(role.GetName(), "role1");
  role.SetName("role2");
  EXPECT_EQ(role.GetName(), "role2");
}

TEST(test_role, role_description) {
  auto role1 = Role("role1", "description1");
  EXPECT_EQ(role1.GetName(), "role1");
  EXPECT_EQ(role1.GetDescription(), "description1");
  role1.SetName("role3");
  EXPECT_EQ(role1.GetDescription(), "description1");

  auto role2 = Role("role2", "description1");
  EXPECT_EQ(role2.GetDescription(), "description1");
  EXPECT_EQ(role2.GetName(), "role2");
  role2.SetDescription("description2");
  EXPECT_EQ(role2.GetDescription(), "description2");
  EXPECT_EQ(role2.GetName(), "role2");
}

TEST(test_role, grant_revoke) {
  auto role1 = Role("role1", "description1");
  EXPECT_TRUE(role1.GetPermissionRules().empty());
  EXPECT_EQ(role1.ToString(), "{\n \"name\": \"role1\",\n \"description\": \"description1\"\n}\n");

  role1.Grant("SOME_PERMISSION", false);
  EXPECT_FALSE(role1.GetPermissionRules().empty());

  EXPECT_EQ(role1.ToString(),
            "{\n \"name\": \"role1\",\n \"description\": \"description1\",\n \"permissionRules\": "
            "[\n  {\n   \"permission\": \"SOME_PERMISSION\",\n   \"grant\": true\n  }\n ]\n}\n");

  role1.Grant("SOME_OTHER_PERMISSION", true);
  EXPECT_EQ(role1.GetPermissionRules().size(), 2);

  role1.Grant("SOME_OTHER_PERMISSION", true);
  // is a set
  EXPECT_EQ(role1.GetPermissionRules().size(), 2);

  role1.RevokeGrant("SOME_PERMISSION", false);
  EXPECT_FALSE(role1.GetPermissionRules().empty());
  role1.RevokeGrant("NOT_EXISTING_PERMISSION", true);

  // is immutable
  for (const auto &rule : role1.GetPermissionRules()) {
    EXPECT_EQ(rule.GetPermission(), "SOME_OTHER_PERMISSION");
  }

  role1.RevokeGrant("SOME_OTHER_PERMISSION", true);
  EXPECT_TRUE(role1.GetPermissionRules().empty());

  EXPECT_EQ(role1.ToString(), "{\n \"name\": \"role1\",\n \"description\": \"description1\"\n}\n");
}

TEST(test_role, deny_revoke) {
  auto role1 = Role("role1", "description1");
  EXPECT_TRUE(role1.GetPermissionRules().empty());
  EXPECT_EQ(role1.ToString(), "{\n \"name\": \"role1\",\n \"description\": \"description1\"\n}\n");

  role1.Deny("SOME_PERMISSION", false);
  EXPECT_FALSE(role1.GetPermissionRules().empty());

  EXPECT_EQ(role1.ToString(),
            "{\n \"name\": \"role1\",\n \"description\": \"description1\",\n \"permissionRules\": "
            "[\n  {\n   \"permission\": \"SOME_PERMISSION\"\n  }\n ]\n}\n");

  role1.Deny("SOME_OTHER_PERMISSION", true);
  EXPECT_EQ(role1.GetPermissionRules().size(), 2);

  role1.Deny("SOME_OTHER_PERMISSION", true);
  // is a set
  EXPECT_EQ(role1.GetPermissionRules().size(), 2);

  role1.RevokeDenial("SOME_PERMISSION", false);
  EXPECT_FALSE(role1.GetPermissionRules().empty());
  role1.RevokeDenial("NOT_EXISTING_PERMISSION", true);

  // is immutable
  for (const auto &rule : role1.GetPermissionRules()) {
    EXPECT_EQ(rule.GetPermission(), "SOME_OTHER_PERMISSION");
  }

  role1.ClearPermissionRules();
  EXPECT_TRUE(role1.GetPermissionRules().empty());

  EXPECT_EQ(role1.ToString(), "{\n \"name\": \"role1\",\n \"description\": \"description1\"\n}\n");
}

TEST(test_role, permission_rule) {
  auto rule = PermissionRule("SOME_PERMISSION", true, false);
  EXPECT_EQ(rule.GetPermission(), "SOME_PERMISSION");
  EXPECT_TRUE(rule.IsGrant());
  EXPECT_FALSE(rule.IsPriority());
  EXPECT_EQ(rule.ToString(), "{\n \"permission\": \"SOME_PERMISSION\",\n \"grant\": true\n}\n");
}

} // namespace caosdb::acm
